#include <exec/devices.h>
#include <exec/io.h>
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/resident.h>
#include <exec/semaphores.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <libraries/expansion.h>
#include <libraries/configvars.h>
#include <workbench/workbench.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/icon.h>
#include <proto/expansion.h>

#if defined(__MORPHOS__) && (__GNUC__ == 2)
void memclr(APTR, ULONG);
#endif

#ifndef __SASC
#define __aligned __attribute__((aligned(4)))
#else
typedef unsigned long	IPTR;
#endif
#include <resources/filesysres.h>
#include <string.h>
#include <stdlib.h>

#ifndef __SASC
# define   DEBUG 1
# include  <aros/debug.h>
#endif

#define	DEBUG_SCAN(x)	;

#include "binddrivers_version.h"

static const char version[] = VERSTAG;

struct ExecBase		*SysBase;
struct DosLibrary	*DOSBase;

struct Library	*IconBase;
struct Library	*ExpansionBase;

char            dirname[1024];

struct Resident* FindLibResident(BPTR	Segment);
struct ConfigDev	*GetConfigDev(char *String);

int main(void)
{
	UBYTE stack_ap[sizeof(struct AnchorPath) + 3];
	struct AnchorPath *MyAp = (struct AnchorPath *) (((ULONG) stack_ap + 3) & ~3);
	struct DiskObject 	*MyDiskObject;
	char			*Product;
	struct ConfigDev	*MyConfigDev;
	BPTR			MySegment;
	struct Resident		*MyResident;
	ULONG                   retval = RETURN_ERROR;
	LONG			error = RETURN_FAIL;


	SysBase=*((struct ExecBase**) 4);

	if ((DOSBase=(struct DosLibrary *)OpenLibrary("dos.library",37))!=0)
	{
		if ((IconBase=(struct Library *)OpenLibrary("icon.library",37)))
		{
			if ((ExpansionBase=OpenLibrary("expansion.library",37)))
			{
				memclr(MyAp,
					   sizeof(*MyAp));

				error = RETURN_OK;

				ObtainConfigBinding();

				for (retval = MatchFirst("SYS:Expansion/#?.info",MyAp);
						retval == 0;
						retval = MatchNext(MyAp))
				{
					if (MyAp->ap_Flags & APF_DirChanged)
					{
						DEBUG_SCAN(dprintf("ScanDrivers: Changed directories...\n"));
					}
					if (NameFromLock(MyAp->ap_Current->an_Lock,
									 dirname,
									 sizeof(dirname)) == FALSE)
					{
						PrintFault(IoErr(), "Error on NameFromLock");
						error = RETURN_WARN;
						break;
					}
					if (AddPart(dirname,&(MyAp->ap_Info.fib_FileName[0]),sizeof(dirname)) == FALSE)
					{
						PrintFault(IoErr(), "Error on AddPart");
						error = RETURN_WARN;
						break;
					}
					//          DEBUG_SCAN(dprintf("ScanDrivers: Check <%s>\n",dirname));

					/* This code section deals with ALL and directory ascent/descent */
					if (MyAp->ap_Info.fib_DirEntryType > 0)
					{
						if (MyAp->ap_Flags & APF_DIDDIR)
						{
							DEBUG_SCAN(dprintf("ScanDrivers: Ascending from directory <%s>\n",(ULONG)dirname));
						}
#if 0
						else
							if (MyArgs.All)
							{
								DEBUG_SCAN(dprintf("ScanDrivers: Entering directory <%s>\n",dirname));

								/* make it enter the directory */
								MyAp->ap_Flags   |=      APF_DODIR;
							}
#endif
							else
							{
								DEBUG_SCAN(dprintf("ScanDrivers: The next dir is  ... <%s>\n",(ULONG)dirname));
							}

						/* clear the completed directory flag */
						MyAp->ap_Flags     &=      ~APF_DIDDIR;

					}
					else
					{
						/* Here is code for handling each particular file */
						DEBUG_SCAN(dprintf("ScanDrivers: Check <%s>\n",(ULONG)dirname));

						if ((MyDiskObject = GetDiskObjectNew(dirname)))
						{
							DEBUG_SCAN(dprintf("ScanDrivers: DiskObject 0x%lx\n",(ULONG)MyDiskObject));
							if ((Product=FindToolType(MyDiskObject->do_ToolTypes,"PRODUCT")))
							{
								DEBUG_SCAN(dprintf("ScanDrivers: Product <%s>\n",(ULONG)Product));
								if ((MyConfigDev=GetConfigDev(Product)))
								{
									//char *Ptr;
									DEBUG_SCAN(dprintf("ScanDrivers: ConfigDev 0x%lx\n",(ULONG)MyConfigDev));
									dirname[strlen(dirname)-sizeof(".info")] = 0;
									DEBUG_SCAN(dprintf("ScanDrivers: Load <%s>\n",(ULONG)dirname));
									if ((MySegment=LoadSeg(dirname)))
									{
										if ((MyResident=FindLibResident(MySegment)))
										{
											struct CurrentBinding MyBinding;
											MyBinding.cb_ConfigDev	=	MyConfigDev;
											MyBinding.cb_FileName	=	dirname;
											MyBinding.cb_ToolTypes	=(UBYTE**) MyDiskObject->do_ToolTypes;
											MyBinding.cb_ProductString=	Product;
											SetCurrentBinding(&MyBinding,sizeof(MyBinding));
											if (InitResident(MyResident,MySegment))
											{
												DEBUG_SCAN(dprintf("ScanDrivers: InitResident worked\n"));
											}
											else
											{
												DEBUG_SCAN(dprintf("ScanDrivers: InitResident failed\n"));
												UnLoadSeg(MySegment);
											}
										}
										else
										{
											UnLoadSeg(MySegment);
										}
									}
								}
							}
							FreeDiskObject(MyDiskObject);
						}
						else
						{
							DEBUG_SCAN(dprintf("ScanDrivers: no diskobject\n"));
						}
					}
				}
				MatchEnd(MyAp);

				ReleaseConfigBinding();

				CloseLibrary(ExpansionBase);
			}
			CloseLibrary((struct Library *)IconBase);
		}
		CloseLibrary((struct Library *)DOSBase);
	}
	DEBUG_SCAN(dprintf("ScanDrivers: ret %ld\n",error));
	return (error);
}

#warning "needs to be tested"

struct ConfigDev	*GetConfigDev(char *String)
{
	struct ConfigDev	*MyConfigDev=NULL;
	struct ConfigDev	*CurrentConfigDev;
	ULONG			ManuID;
	ULONG			ProductID;
	char			*NewString;

	if ((String=strchr(String,(ULONG)"=")))
	{
		/*
		 * skip "="
		 */
		String++;
	}

	while (String && *String)
	{
		ManuID = atoi(String);
		if ((NewString=strchr(String,(ULONG)"/")))
		{
			/*
			 * skip "/"
			 */
			NewString++;
			ProductID = atoi(NewString);
			String = NewString;
		}
		else
		{
			/*
			 * only a manufactorer id
			 */
			ProductID = -1;
		}

		if ((String=strchr(String,(ULONG)"|")))
		{
			/*
			 * skip "|"
			 */
			String++;
		}
		else
		{
			/*
			 * no further element
			 */
		}

		CurrentConfigDev = NULL;
		while ((CurrentConfigDev=FindConfigDev(CurrentConfigDev,
											   ManuID,
											   ProductID)))
		{
			CurrentConfigDev->cd_NextCD = MyConfigDev;
			MyConfigDev = CurrentConfigDev;
		}
	}
	return(MyConfigDev);
}

struct Resident* FindLibResident(BPTR	Segment)
{
	struct Resident *MyResident;
	ULONG		*MySegment;
	UWORD		*MyBuffer;
	UWORD		*EndBuffer;

	while (Segment)
	{
		MySegment	=(ULONG*) BADDR(Segment);
		MyBuffer	=(UWORD*) &MySegment[1];
		EndBuffer	=(UWORD*) &MySegment[MySegment[-1]/sizeof(ULONG)];
		while (MyBuffer < EndBuffer)
		{
			if (*MyBuffer==RTC_MATCHWORD)
			{
				MyResident=(struct Resident*) MyBuffer;
				if (MyResident->rt_MatchTag == MyResident)
				{
					return(MyResident);
				}
			}
			MyBuffer++;
		}
		Segment	=(BPTR) MySegment[0];
	}
	return(NULL);
}

